from __future__ import annotations
from tuneflow_py.models.protos import song_pb2


class Note:
    def __init__(self, proto: song_pb2.Note | None = None) -> None:
        self._proto = proto if proto is not None else song_pb2.Note()

    def get_pitch(self) -> int:
        return self._proto.pitch

    def get_velocity(self) -> int:
        return self._proto.velocity

    def get_start_tick(self) -> int:
        return self._proto.start_tick

    def get_end_tick(self) -> int:
        return self._proto.end_tick

    def set_pitch(self, pitch: int):
        if not Note.is_valid_pitch(pitch):
            raise Exception("Invalid note pitch " + str(pitch))
        self._proto.pitch = pitch

    def get_start_time(self) -> float:
        return self._proto.start_time

    def get_end_time(self) -> float:
        return self._proto.end_time

    @staticmethod
    def is_valid_pitch(pitch: int):
        return pitch >= 0 and pitch <= 127 and isinstance(pitch, int)



from __future__ import annotations
from tuneflow_py.descriptors.text import LabelText
from tuneflow_py.descriptors.common import RealNumber
from enum import Enum
from typing import TypedDict, Optional, Any, Literal
from tuneflow_py.models.track import TrackType


class WidgetType(Enum):
    Slider = 1
    Input = 2
    # A widget that selects one track.
    TrackSelector = 3
    # A widget that selects a pitch.
    Pitch = 4
    TrackPitchSelector = 5
    InstrumentSelector = 6
    Select = 7
    Switch = 8
    InputNumber = 9
    MultiTrackSelector = 10
    # Does not render a widget.
    NoWidget = 11
    FileSelector = 12
    MultiSourceAudioSelector = 13
    AudioRecorder = 14
    SelectList = 15
    # Read-only table of descriptions.
    Descriptions = 16


class SliderWidgetConfig(TypedDict, total=False):
    min_value: RealNumber
    max_value: RealNumber
    step: RealNumber
    unit: Optional[str]


class InputWidgetConfig(TypedDict, total=False):
    min_value: RealNumber
    max_value: RealNumber
    step: RealNumber


class TrackSelectorWidgetConfig(TypedDict, total=False):
    always_show_track_info: Optional[bool]
    '''  Whether to always show the track info. Default to false. '''

    allowed_track_types: Optional[list[TrackType]]  # type: ignore
    ''' If specified, only the allowed types of tracks can be selected. '''


class PitchWidgetConfig(TypedDict, total=False):
    min_allowed_pitch: Optional[RealNumber]
    max_allowed_pitch: Optional[RealNumber]


class TrackPitchSelectorWidgetConfig(TypedDict, total=False):
    track_selector_config: TrackSelectorWidgetConfig
    pitch_selector_config: PitchWidgetConfig


class InstrumentSelectorWidgetConfig(TypedDict, total=False):
    disabled_programs: Optional[list[RealNumber]]
    ''' Not supported yet. '''


class SelectWidgetOption(TypedDict, total=False):
    label: LabelText
    value: Any


class SelectWidgetConfig(TypedDict, total=False):
    options: list[SelectWidgetOption]

    allow_search: Optional[bool]
    ''' Whether to show search box. Default to false. '''

    placeholder: LabelText

    virtual_list_props: Optional[Any]
    ''' https://arco.design/vue/component/select#virtual-list '''

    populate_options_with_generatable_styles: Optional[bool]
    '''
    Whether to populate the options with the styles that
    TuneFlow can generate.
    '''

    populate_options_with_generatable_tempos: Optional[bool]
    '''
    Whether to populate the options with the tempo settings that
    TuneFlow can generate with.
    '''


class SelectListWidgetConfig(TypedDict, total=False):
    options: list[SelectWidgetOption]
    max_height: Optional[RealNumber]
    size: Optional[str]
    virtual_list_props: Optional[Any]
    ''' https://arco.design/vue/component/list '''

    allow_search: Optional[bool]


class SwitchWidgetConfig(TypedDict, total=False):
    type: Optional[Literal['circle', 'round', 'line']]
    '''
    'circle' | 'round' | 'line'
    
    https://arco.design/vue/component/switch
    '''


class InputNumberWidgetConfig(TypedDict, total=False):
    min_value: RealNumber
    max_value: RealNumber
    step: RealNumber


class FileSelectorWidgetConfig(TypedDict, total=False):
    allowed_extensions: list[str]
    ''' The extensions (without ".") that are allowed to choose. '''

    select_directory: Optional[str]
    '''
    Whether to select a directory instead of a file.
    Default to false.
    '''

    placeholder: Optional[LabelText]
    ''' Custom placeholder text. '''

    select_local_file: Optional[bool]
    ''' If true, selects local system files. '''


AudioSourceType = Literal['file', 'audioTrack', 'record']


class MultiSourceAudioSelectorWidgetConfig(TypedDict, total=False):
    allowed_sources: Optional[list[AudioSourceType]]
    ''' Default to allow all audio sources. '''


class MultiSourceAudioSelectorResult(TypedDict, total=False):
    source_type: AudioSourceType

    audio_info: Any
    '''
    Result type will be:
    * `File` if `sourceType` is 'file'
    * trackId if `sourceType` is 'audioTrack'
    * `AudioBuffer` if `sourceType` is 'record'
    '''


class DescriptionData(TypedDict, total=False):
    label: LabelText
    value: str
    span: RealNumber


class DescriptionsWidgetConfig(TypedDict, total=False):
    size: Literal['mini', 'small', 'medium', 'large']
    column: int
    data: list[DescriptionData]


class WidgetDescriptor(TypedDict, total=False):
    type: int  # WidgetType
    config: Optional[SliderWidgetConfig
                     | InputWidgetConfig
                     | SelectWidgetConfig
                     | TrackSelectorWidgetConfig
                     | PitchWidgetConfig
                     | TrackPitchSelectorWidgetConfig
                     | InstrumentSelectorWidgetConfig
                     | SwitchWidgetConfig
                     | InputNumberWidgetConfig
                     | FileSelectorWidgetConfig
                     | SelectListWidgetConfig
                     | MultiSourceAudioSelectorWidgetConfig
                     | DescriptionsWidgetConfig]



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2

AutomationTargetType = song_pb2.AutomationTarget.TargetType


class AutomationTarget:
    def __init__(self, type: AutomationTargetType | None = None, plugin_instance_id: str | None = None, param_id: str | None = None, proto: song_pb2.AutomationTarget | None = None) -> None:
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.AutomationTarget(
                type=type, plugin_instance_id=plugin_instance_id, param_id=param_id)

    def get_type(self):
        return self._proto.type

    def set_type(self, type: AutomationTargetType):
        self._proto.type = type

    def get_plugin_instance_id(self):
        return self._proto.audio_plugin_id

    def set_plugin_instance_id(self, plugin_instance_id: str | None = None):
        if plugin_instance_id is None:
            del self._proto.audio_plugin_id
        else:
            self._proto.audio_plugin_id = plugin_instance_id

    def get_param_id(self):
        return self._proto.param_id

    def set_param_id(self, param_id: str | None = None):
        if param_id is None:
            del self._proto.param_id
        else:
            self._proto.param_id = param_id

    def to_tf_automation_target_id(self):
        '''
        Gets a unique id that identifies this target type.
        '''
        return AutomationTarget.encode_automation_target(self.get_type(), self.get_plugin_instance_id(), self.get_param_id())

    @staticmethod
    def encode_automation_target(
        target_type: AutomationTargetType,
        plugin_instance_id: str | None,
        param_id: str | None,
    ):
        if target_type == AutomationTargetType.AUDIO_PLUGIN:
            return f'{target_type}^^{plugin_instance_id}^^{param_id}'

        return f'{target_type}'

    @staticmethod
    def decode_automation_target(encoded_target: str):
        parts = encoded_target.split('^^')
        if len(parts) == 0:
            raise Exception(f'Invalid automation target id: {encoded_target}')

        type = int(parts[0])
        if len(parts) > 2:
            AutomationTarget(type, parts[1], parts[2])

        return AutomationTarget(type)



from __future__ import annotations
from base64 import b64encode, b64decode
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.track import Track, TrackType
from tuneflow_py.models.clip import Clip, ClipType
from tuneflow_py.models.tempo import TempoEvent
from tuneflow_py.models.time_signature import TimeSignatureEvent
from tuneflow_py.models.automation import AutomationTarget, AutomationTargetType
from tuneflow_py.utils import db_to_volume_value, greater_equal, lower_than
from miditoolkit.midi import MidiFile, TempoChange as ToolkitTempoChange, TimeSignature as ToolkitTimeSignature, Instrument, Note as ToolkitNote
from types import SimpleNamespace


class Song:
    def __init__(self, proto: song_pb2.Song | None = None) -> None:
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.Song()
            self._proto.PPQ = Song.get_default_resolution()
            self._proto.tempos.append(
                song_pb2.TempoEvent(ticks=0, time=0, bpm=120))
            self._proto.time_signatures.append(
                song_pb2.TimeSignatureEvent(ticks=0, numerator=4, denominator=4))
            self._proto.last_tick = 0
            self._proto.duration = 0
            self._proto.master_track.uuid = Track._generate_track_id()
            self._proto.master_track.type = TrackType.MASTER_TRACK
            self._proto.master_track.volume = db_to_volume_value(0.0)

    def get_last_tick(self):
        '''
        @returns End tick of the last note.
        '''
        if self.get_track_count() == 0:
            return 0
        return max([track.get_track_end_tick() for track in self.get_tracks()])

    def get_duration(self):
        return self.tick_to_seconds(self.get_last_tick())

    def get_track_count(self):
        return len(self._proto.tracks)

    def get_tracks(self):
        for track_proto in self._proto.tracks:
            yield Track(song=self, proto=track_proto)

    def get_track_by_id(self, track_id: str) -> Track | None:
        for track_proto in self._proto.tracks:
            if track_proto.uuid == track_id:
                return Track(song=self, proto=track_proto)
        return None

    def get_track_at(self, index):
        return Track(song=self, proto=self._proto.tracks[index])

    def serialize(self):
        # type: ignore
        return b64encode(self._proto.SerializeToString()).decode('ascii')

    @staticmethod
    def deserialize(serialized_song_string: str):
        song_proto = song_pb2.Song()
        song_proto.ParseFromString(
            b64decode(serialized_song_string))  # type: ignore
        return Song(proto=song_proto)

    @staticmethod
    def from_midi(midi_obj: MidiFile):
        '''
        TODO: Replace proto operations with builtin methods.
        '''
        def scale_int_by(value, scale_factor):
            return round(value*scale_factor)
        song = Song()
        song_proto = song._proto
        ppq_scale_factor = float(song_proto.PPQ) / \
            float(midi_obj.ticks_per_beat)
        # Add tempos and time signatures
        song.overwrite_tempo_changes([TempoEvent(ticks=scale_int_by(
            tempo_change.time, ppq_scale_factor), bpm=tempo_change.tempo) for tempo_change in midi_obj.tempo_changes])
        del song_proto.time_signatures[:]
        for time_signature_change in midi_obj.time_signature_changes:
            song_proto.time_signatures.add(ticks=scale_int_by(
                time_signature_change.time, ppq_scale_factor), numerator=time_signature_change.numerator, denominator=time_signature_change.denominator)

        # Add tracks and notes.
        song_last_tick = 0
        for index, instrument in enumerate(midi_obj.instruments):
            song_track_proto = song_proto.tracks.add(
                uuid=Track._generate_track_id(), rank=index, type=TrackType.MIDI_TRACK)
            song_track_proto.instrument.program = instrument.program
            song_track_proto.instrument.is_drum = instrument.is_drum
            track_clip_proto = song_track_proto.clips.add(
                id=Clip._generate_clip_id(), type=ClipType.MIDI_CLIP, clip_start_tick=0)
            # Add notes.
            for note in instrument.notes:
                start_tick = scale_int_by(
                    note.start, ppq_scale_factor)
                end_tick = scale_int_by(note.end, ppq_scale_factor)
                track_clip_proto.notes.add(pitch=note.pitch, velocity=note.velocity, start_tick=start_tick, start_time=song.tick_to_seconds(
                    start_tick), end_tick=end_tick, end_time=song.tick_to_seconds(end_tick))
            track_clip_proto.clip_start_tick = min(
                track_clip_proto.notes, key=lambda x: x.start_tick).start_tick
            track_clip_proto.clip_end_tick = max(
                track_clip_proto.notes, key=lambda x: x.end_tick).end_tick
            song_last_tick = max(
                song_last_tick, track_clip_proto.clip_end_tick)
            # Add automation.
            volume_ccs = []
            pan_ccs = []
            for control_change in instrument.control_changes:
                if control_change.number == 7:
                    # Volume CC
                    volume_ccs.append(control_change)
                elif control_change.number == 10:
                    # Pan CC
                    pan_ccs.append(control_change)
            if len(volume_ccs) == 1:
                song_track_proto.volume = volume_ccs[0].value/127.0
            elif len(volume_ccs) > 1:
                volume_target = AutomationTarget(AutomationTargetType.VOLUME)
                volume_target_id = volume_target.to_tf_automation_target_id()
                song_track_proto.automation.targets.append(
                    volume_target._proto)
                volume_target_value = song_pb2.AutomationValue()
                song_track_proto.automation.target_values[volume_target_id] = volume_target_value
                for index, cc in enumerate(sorted(volume_ccs, key=lambda x: x.time)):
                    volume_target_value.points.add(tick=scale_int_by(
                        cc.time, ppq_scale_factor), value=cc.value/127.0, id=index+1)
            if len(pan_ccs) == 1:
                song_track_proto.pan = pan_ccs[0].value - 64
            elif len(pan_ccs) > 1:
                pan_target = AutomationTarget(AutomationTargetType.PAN)
                pan_target_id = pan_target.to_tf_automation_target_id()
                song_track_proto.automation.targets.append(pan_target._proto)
                pan_target_value = song_pb2.AutomationValue()
                song_track_proto.automation.target_values[pan_target_id] = pan_target_value
                for index, cc in enumerate(sorted(pan_ccs, key=lambda x: x.time)):
                    pan_target_value.points.add(tick=scale_int_by(
                        cc.time, ppq_scale_factor), value=cc.value/127.0, id=index+1)

        song.last_tick = song_last_tick
        song.duration = song.tick_to_seconds(song_last_tick)
        return song

    def to_midi(self):
        '''
        TODO: Replace proto operations with builtin methods.
        '''
        midi_obj = MidiFile()
        midi_obj.ticks_per_beat = self.get_resolution()
        for tempo_proto in self._proto.tempos:
            midi_obj.tempo_changes.append(ToolkitTempoChange(
                tempo=tempo_proto.bpm, time=tempo_proto.ticks))
        for time_signature_proto in self._proto.time_signatures:
            midi_obj.time_signature_changes.append(ToolkitTimeSignature(
                numerator=time_signature_proto.numerator, denominator=time_signature_proto.denominator, time=time_signature_proto.ticks))
        for track_proto in self._proto.tracks:
            if track_proto.type != TrackType.MIDI_TRACK or len(track_proto.clips) == 0:
                continue
            instrument = Instrument(program=track_proto.instrument.program,
                                    is_drum=track_proto.instrument.is_drum, name=f'Track {track_proto.rank}')
            midi_obj.instruments.append(instrument)
            # Export clips
            for clip_proto in track_proto.clips:
                if clip_proto.type != ClipType.MIDI_CLIP:
                    continue
                for note_proto in clip_proto.notes:
                    if not Clip.is_note_in_clip(note_start_tick=note_proto.start_tick, note_end_tick=note_proto.end_tick, clip_start_tick=clip_proto.clip_start_tick, clip_end_tick=clip_proto.clip_end_tick):
                        continue
                    instrument.notes.append(ToolkitNote(
                        pitch=note_proto.pitch, velocity=note_proto.velocity, start=note_proto.start_tick, end=note_proto.end_tick))
            # TODO: Export automation
        return midi_obj

    def get_resolution(self):
        return self._proto.PPQ

    def get_tempo_event_count(self):
        return len(self._proto.tempos)

    def create_tempo_change(self, ticks: int, bpm: float):
        '''
        Adds a tempo change event into the song and returns it.

        @param ticks The tick at which this event happens.
        @param bpm The new tempo in BPM(Beats-per-minute) format.
        '''
        if self.get_resolution() <= 0:
            raise Exception(
                'Song resolution must be provided before creating tempo changes.')

        if self.get_tempo_event_count() == 0 and ticks != 0:
            raise Exception('The first tempo event must be at tick 0')

        # Calculate time BEFORE the new tempo event is inserted.
        tempo_change = TempoEvent(
            ticks=ticks, bpm=bpm, time=self.tick_to_seconds(ticks))
        insert_index = greater_equal(
            self._proto.tempos,
            tempo_change._proto,
            lambda x: x.ticks
        )
        if insert_index < 0:
            self._proto.tempos.append(tempo_change._proto)
        else:
            self._proto.tempos.insert(insert_index, tempo_change._proto)

        self.retiming_tempo_events()
        return tempo_change

    def retiming_tempo_events(self):
        sorted_tempos = sorted(
            self._proto.tempos, key=lambda tempo: tempo.ticks)
        del self._proto.tempos[:]
        self._proto.tempos.extend(sorted_tempos)
        # Re-calculate all tempo event time.
        for tempo_event_proto in self._proto.tempos:
            tempo_event_proto.time = self.tick_to_seconds(
                tempo_event_proto.ticks)

    def tick_to_seconds(self, tick: int):
        if tick == 0:
            return 0

        target_tempo = SimpleNamespace()
        target_tempo.ticks = tick
        base_tempo_index = lower_than(
            self._proto.tempos,
            target_tempo,
            lambda x: x.ticks
        )
        if base_tempo_index == -1:
            # If no tempo is found before the tick, use the first tempo.
            base_tempo_index = 0

        base_tempo_change = self._proto.tempos[base_tempo_index]
        ticks_delta = tick - base_tempo_change.ticks
        ticks_per_second_since_last_tempo_change = Song._tempo_bpm_to_ticks_per_second(
            base_tempo_change.bpm,
            self.get_resolution(),
        )
        return base_tempo_change.time + ticks_delta / ticks_per_second_since_last_tempo_change

    def overwrite_tempo_changes(self, tempo_events: list[TempoEvent]):
        if len(tempo_events) == 0:
            raise Exception('Cannot clear all the tempo events.')
        sorted_tempo_events = sorted(
            tempo_events, key=lambda tempo_event: tempo_event.get_ticks())
        first_tempo_event = sorted_tempo_events[0]
        if first_tempo_event.get_ticks() > 0:
            raise Exception('The first tempo event needs to start from tick 0')
        del self._proto.tempos[:]
        self._proto.tempos.append(song_pb2.TempoEvent(
            ticks=0, time=0, bpm=first_tempo_event.get_bpm()))
        for i in range(1, len(sorted_tempo_events)):
            tempo_event = sorted_tempo_events[i]
            self.create_tempo_change(
                ticks=tempo_event.get_ticks(), bpm=tempo_event.get_bpm())
        self.retiming_tempo_events()

    def get_time_signature_event_count(self):
        return len(self._proto.time_signatures)

    def get_time_signature_event_at(self, index: int):
        return TimeSignatureEvent(proto=self._proto.time_signatures[index])

    def create_track(self, type: int,
                     index: int,
                     rank: int | None = None):
        new_track = Track(
            type=type, song=self, rank=rank if rank is not None else self.get_next_track_rank())
        self._proto.tracks.insert(index, new_track._proto)
        return new_track

    def get_next_track_rank(self):
        return 1 if len(self._proto.tracks) == 0 else max([track.rank for track in self._proto.tracks]) + 1

    @staticmethod
    def _tempo_bpm_to_ticks_per_second(tempo_bpm: float, PPQ: int):
        return (tempo_bpm * PPQ) / 60

    @staticmethod
    def get_default_resolution():
        '''
        Returns the default Pulse-per-Quater-Note used in TuneFlow.
        '''
        return 480



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.clip import Clip
from tuneflow_py.utils import db_to_volume_value, volume_value_to_db
import nanoid
TrackType = song_pb2.TrackType


class Track:
    def __init__(self, type: int | None = None,
                 song=None,
                 uuid=None,
                 instrument=None,
                 volume=db_to_volume_value(0),
                 solo=False,
                 muted=False,
                 rank=0,
                 pan=0, proto: song_pb2.Track | None = None) -> None:
        if song is None:
            raise Exception('song must be provided when creating track')
        self.song = song
        if proto is not None:
            self._proto = proto
            return
        self._proto = song_pb2.Track()
        self._proto.type = type
        if instrument is not None:
            self._proto.instrument.Merge(instrument)
        elif type == TrackType.MIDI_TRACK:
            self._proto.instrument.program = 0
            self._proto.instrument.is_drum = False
        self._proto.uuid = uuid if uuid is not None else Track._generate_track_id()
        self._proto.volume = volume
        self._proto.solo = solo
        self._proto.muted = muted
        self._proto.rank = rank
        self._proto.pan = pan

    def get_clip_count(self):
        return len(self._proto.clips)

    def get_clips(self):
        for clip_proto in self._proto.clips:
            yield Clip(proto=clip_proto)

    def get_clip_at(self, clip_index: int):
        return Clip(proto=self._proto.clips[clip_index])

    def get_volume(self):
        '''
        Returns the fader position, ranging from 0 to 1.
        '''
        return self._proto.volume

    def get_volume_in_db(self):
        return volume_value_to_db(self._proto.volume)

    def get_pan(self):
        '''
        A pan value from -64 to 63
        '''
        return self._proto.pan

    def get_instrument(self):
        return self._proto.instrument

    def get_track_start_tick(self):
        if self.get_clip_count() == 0:
            return 0
        return self.get_clip_at(0).get_clip_start_tick()

    def get_track_end_tick(self):
        if self.get_clip_count() == 0:
            return 0
        return self.get_clip_at(self.get_clip_count() - 1).get_clip_end_tick()

    @staticmethod
    def _generate_track_id():
        return nanoid.generate()



from typing import Literal, TypedDict, Optional, Union

RealNumber = Union[int, float]


class ClipInfo(TypedDict):
    track_id: str
    clip_id: str


SupportedPlatform = Literal['desktop', 'web']


class PluginInfo(TypedDict, total=False):
    supported_platforms: Optional[SupportedPlatform]
    ''' If not provided, supports all platforms. '''

    min_required_desktop_version: Optional[str]
    ''' If specified, the desktop requires a minimum version to run this plugin. '''

    is_in_development: Optional[bool]
    ''' If true, this plugin is in development stage and can only be accessed by its developer. '''



from typing import TypedDict, Optional


class SongAccess(TypedDict, total=False):
    create_track: Optional[bool]
    ''' Whether the plugin has permission to create a new track. '''

    remove_track: Optional[bool]
    ''' Whether the plugin has permission to remove a track. '''



from __future__ import annotations
from tuneflow_py.base_plugin import TuneflowPlugin, ReadAPIs
from tuneflow_py.models.automation import AutomationTarget, AutomationTargetType
from tuneflow_py.models.clip import ClipType, Clip
from tuneflow_py.models.note import Note
from tuneflow_py.models.song import Song
from tuneflow_py.models.tempo import TempoEvent
from tuneflow_py.models.time_signature import TimeSignatureEvent
from tuneflow_py.models.track import TrackType, Track
from tuneflow_py.descriptors.widget import *
from tuneflow_py.descriptors.text import *
from tuneflow_py.descriptors.param import *
from tuneflow_py.descriptors.song_access import *
from tuneflow_py.descriptors.common import *
